@@ -8,10 +8,11 @@ import worlds._bizhawk as bizhawk
 from worlds._bizhawk.client import BizHawkClient
 from .utils import Constants
 from .items import is_dice_item, convert_item_id_to_dice_id, item_id_to_item_name
-from .locations import get_location_id_for_duelist, duelist_from_location_id, is_duelist_location_id, get_location_id_for_duelist_rematch, duelist_rematch_from_location_id, is_duelist_rematch_location_id, get_location_id_for_tournament
+from .locations import get_location_id_for_duelist, duelist_from_location_id, is_duelist_location_id, get_2nd_location_id_for_duelist, get_location_id_for_tournament, get_2nd_location_id_for_tournament, get_3rd_location_id_for_tournament, get_location_id_for_dice_id
 from .duelists import Duelist, all_duelists, name_to_duelist
 from .dice import id_to_dice
 from .tournament import Tournament, all_tournaments
+from .options import BonusItemMode
 from .version import __version__
 
 if TYPE_CHECKING:
@@ -121,6 +122,69 @@ class YGODDMClient(BizHawkClient):
                 COMBINED_WRAM
             )])
 
+    async def award_gold(self, ctx: "BizHawkClientContext", number_to_award: int) -> bool:
+        current_money_bytes = (await bizhawk.read(
+            ctx.bizhawk_ctx, [(Constants.MONEY_OFFSET, 2, COMBINED_WRAM)]
+        ))[0]
+        current_money: int = int.from_bytes(current_money_bytes, "little", signed = True)
+        m1, m2 = (current_money & 0xFFFF).to_bytes(2, "little")
+        current_money = current_money + (number_to_award * ctx.slot_data[Constants.GAME_OPTIONS_KEY]["gold_reward_amount"])
+        if (current_money > 65000):
+            current_money = 65000
+        g1, g2 = (current_money & 0xFFFF).to_bytes(2, "little")
+        # Guarded write based on read in gold amount. Stops things from messing up when the game is updating gold value
+        award_success: bool = await bizhawk.guarded_write(ctx.bizhawk_ctx, [(
+                Constants.MONEY_OFFSET,
+                [g1, g2],
+                COMBINED_WRAM
+            )],[(
+                Constants.MONEY_OFFSET,
+                [m1, m2],
+                COMBINED_WRAM
+            )])
+        return award_success
+
+    # Sets the shop progress to the appropriate amount based on what has been received
+    async def set_shop_progress(self, ctx: "BizHawkClientContext", progress_amount: int) -> None:
+        # 0 -> 0
+        # 1 -> 3C
+        # 2 -> 69
+        # 3 -> 96
+        # 4 -> D2
+        # 5 -> FF
+        # 6 -> 12C
+        new_shop_prog: int = 0
+        if (progress_amount == 1):
+            new_shop_prog = 0x3C
+        elif (progress_amount == 2):
+            new_shop_prog = 0x69
+        elif (progress_amount == 3):
+            new_shop_prog = 0x96
+        elif (progress_amount == 4):
+            new_shop_prog = 0xD2
+        elif (progress_amount == 5):
+            new_shop_prog = 0xFF
+        elif (progress_amount >= 6):
+            new_shop_prog = 0x12C
+        # Need to overwrite current shop progress so the player can't advance it by playing in tournaments
+        s1, s2 = new_shop_prog.to_bytes(2, "little")
+        await bizhawk.write(ctx.bizhawk_ctx, [(
+                Constants.SHOP_PROGRESS_OFFSET,
+                [s1, s2],
+                COMBINED_WRAM
+            )])
+
+    async def check_dice_collection_locations(self, ctx: "BizHawkClientContext") -> typing.Set[int]:
+        collected_dice: typing.Set[int] = set()
+        dice_collection_memory: bytes = await self.read_dice_collection(ctx)
+        for i in range(len(dice_collection_memory)):
+            if (dice_collection_memory[i] > 0):
+                collected_dice.add(get_location_id_for_dice_id(i)) # count is the dice ID in this case
+        #for dice_amount_byte, count in Counter(dice_collection_memory):
+        #    if (int.from_bytes(dice_amount_byte, "little", signed = True) > 0):
+        #        collected_dice.add[count] # count is the dice ID in this case
+        return collected_dice
+
 
     async def game_watcher(self, ctx: "BizHawkClientContext") -> None:
         if ctx.slot_data is not None:
@@ -168,6 +232,14 @@ class YGODDMClient(BizHawkClient):
                     get_location_id_for_duelist(key) for key, value in duelists_to_wins.items() if value > 0
                 ])
 
+                if (ctx.slot_data[Constants.GAME_OPTIONS_KEY]['free_duel_rewards'] == 1):
+                    # Grab 2nd checks
+                    more_local_check_locations: typing.Set[int] = set([
+                        get_2nd_location_id_for_duelist(key) for key, value in duelists_to_wins.items() if value > 0
+                    ])
+
+                    new_local_check_locations = new_local_check_locations.union(more_local_check_locations)
+
                 # Unlock Duelists
 
                 unlocked_duelist_bitflags: typing.List[int] = await self.read_duelist_collection(ctx)
@@ -199,9 +271,10 @@ class YGODDMClient(BizHawkClient):
                             duelist_bitflag_index = duelist_bitflag_index + 1
                         unlocked_duelist_bitflags[duelist_bitflag_index] |= duelist_bitflag
                 
-                # Check for Yami Yugi unlock based on number of duelists unlocked
-                # If the 91 other duelists are all present, so is Yami Yugi
-                if duelist_count >= 91:
+                # Check for Yami Yugi unlock based on number of duelists defeated
+                # Number defined by yaml option (free_duel_goal)
+                wins_count = sum(value > 0 for value in duelists_to_wins.values())
+                if wins_count >= ctx.slot_data[Constants.GAME_OPTIONS_KEY]['free_duel_goal']:
                     unlocked_duelist_bitflags[0] |= Duelist.YAMI_YUGI.bitflag
 
                 await bizhawk.write(ctx.bizhawk_ctx, [(
@@ -226,6 +299,22 @@ class YGODDMClient(BizHawkClient):
                     get_location_id_for_tournament(key) for key, value in tournaments_to_wins.items() if value
                 ])
 
+                if (ctx.slot_data[Constants.GAME_OPTIONS_KEY]['tournament_rewards'] >= 1):
+                    # Grab 2nd checks
+                    more_local_check_locations: typing.Set[int] = set([
+                        get_2nd_location_id_for_tournament(key) for key, value in tournaments_to_wins.items() if value
+                    ])
+
+                    new_local_check_locations = new_local_check_locations.union(more_local_check_locations)
+
+                if (ctx.slot_data[Constants.GAME_OPTIONS_KEY]['tournament_rewards'] >= 2):
+                    # Grab 3rd checks
+                    more_local_check_locations: typing.Set[int] = set([
+                        get_3rd_location_id_for_tournament(key) for key, value in tournaments_to_wins.items() if value
+                    ])
+
+                    new_local_check_locations = new_local_check_locations.union(more_local_check_locations)
+
                 # Unlock/Lock Tournament Divisions
 
                 division_2_lock: typing.List[int] = [int.from_bytes(tournament_wins_bytes[0], "little") & 0xFE]
@@ -248,43 +337,78 @@ class YGODDMClient(BizHawkClient):
                     COMBINED_WRAM
                 )])
 
-            # Give out received Dice
-            last_dice_received_count: int = int.from_bytes(
-                (await bizhawk.read(ctx.bizhawk_ctx, [(Constants.RECEIVED_DICE_COUNT_OFFSET, 1, COMBINED_WRAM)]))[0]
-            )
+            # Award Dice collection based locations if applicable
+            if (ctx.slot_data[Constants.GAME_OPTIONS_KEY]['bonus_item_mode'] == BonusItemMode.option_shop_progress):
+                dice_locations: typing.Set[int] = await self.check_dice_collection_locations(ctx)
+                new_local_check_locations = new_local_check_locations.union(dice_locations)
+
             received_items: typing.List[int] = [
                 item.item for item in ctx.items_received
             ]
-            received_dice_ids: typing.List[int] = [id for id in received_items if is_dice_item(id)]
 
-            if (len(received_dice_ids) > last_dice_received_count):
-                new_received_dice_ids: typing.List[int] = received_dice_ids[last_dice_received_count:]
-                new_dice_ids: typing.List[int] = [convert_item_id_to_dice_id(i) for i in new_received_dice_ids]
-                # Check if valid Dice id?
+            # Shop Progress
+            if (ctx.slot_data[Constants.GAME_OPTIONS_KEY]['bonus_item_mode'] == BonusItemMode.option_shop_progress):
+                last_shop_prog_received_count: int = int.from_bytes(
+                    (await bizhawk.read(ctx.bizhawk_ctx, [(Constants.RECEIVED_SHOP_PROGRESS_COUNT_OFFSET, 1, COMBINED_WRAM)]))[0]
+                )
 
-                dice_collection_memory: bytes = await self.read_dice_collection(ctx)
-                for dice_id, count in Counter(new_dice_ids).items():
+                received_shop_prog_count: int = sum(1 for item in received_items if item == Constants.SHOP_PROGRESSION_ITEM_ID)
+                
+                if (received_shop_prog_count > last_shop_prog_received_count):
                     await bizhawk.write(ctx.bizhawk_ctx, [(
-                        Constants.DICE_COLLECTION_OFFSET + dice_id,
-                        (dice_collection_memory[dice_id - 1] + count).to_bytes(1, "little"),
-                        COMBINED_WRAM
-                    )])
+                    Constants.RECEIVED_SHOP_PROGRESS_COUNT_OFFSET,
+                    [received_shop_prog_count],
+                    COMBINED_WRAM
+                )])
+                await self.set_shop_progress(ctx, received_shop_prog_count)
 
-                dice_collection_memory: bytes = await self.read_dice_collection(ctx)
+            # Bonus Items (Filler)
 
-                await bizhawk.write(ctx.bizhawk_ctx, [(
-                        Constants.RECEIVED_DICE_COUNT_OFFSET,
-                        len(received_dice_ids).to_bytes(1, "little"),
-                        COMBINED_WRAM
-                    )])
+            #Gold
+            if (ctx.slot_data[Constants.GAME_OPTIONS_KEY]['bonus_item_mode'] == BonusItemMode.option_shop_progress):
+                last_gold_received_count: int = int.from_bytes(
+                    (await bizhawk.read(ctx.bizhawk_ctx, [(Constants.RECEIVED_GOLD_COUNT_OFFSET, 1, COMBINED_WRAM)]))[0]
+                )
 
-            if (ctx.slot_data[Constants.GAME_OPTIONS_KEY]['progression'] == 0 and ctx.slot_data[Constants.GAME_OPTIONS_KEY]['duelist_rematches'] == 1):
-                # Grab rematch checks
-                more_local_check_locations: typing.Set[int] = set([
-                    get_location_id_for_duelist_rematch(key) for key, value in duelists_to_wins.items() if value > 1
-                ])
+                received_gold_count: int = sum(1 for item in received_items if item == Constants.GOLD_FILLER_ITEM_ID)
+                
+                if (received_gold_count > last_gold_received_count):
+                    # Write in new amount of gold bonuses given if the gold is given successfully
+                    award_success: bool = await self.award_gold(ctx, received_gold_count - last_gold_received_count)
+                    if (award_success):
+                        await bizhawk.write(ctx.bizhawk_ctx, [(
+                                Constants.RECEIVED_GOLD_COUNT_OFFSET,
+                                received_gold_count.to_bytes(1, "little"),
+                                COMBINED_WRAM
+                            )])
+                    
 
-                new_local_check_locations = new_local_check_locations.union(more_local_check_locations)
+            # Give out received Dice
+            if (ctx.slot_data[Constants.GAME_OPTIONS_KEY]['bonus_item_mode'] == BonusItemMode.option_random_dice):
+                last_dice_received_count: int = int.from_bytes(
+                    (await bizhawk.read(ctx.bizhawk_ctx, [(Constants.RECEIVED_DICE_COUNT_OFFSET, 1, COMBINED_WRAM)]))[0]
+                )
+                
+                received_dice_ids: typing.List[int] = [id for id in received_items if is_dice_item(id)]
+
+                if (len(received_dice_ids) > last_dice_received_count):
+                    new_received_dice_ids: typing.List[int] = received_dice_ids[last_dice_received_count:]
+                    new_dice_ids: typing.List[int] = [convert_item_id_to_dice_id(i) for i in new_received_dice_ids]
+                    # Check if valid Dice id?
+
+                    dice_collection_memory: bytes = await self.read_dice_collection(ctx)
+                    for dice_id, count in Counter(new_dice_ids).items():
+                        await bizhawk.write(ctx.bizhawk_ctx, [(
+                            Constants.DICE_COLLECTION_OFFSET + dice_id,
+                            (dice_collection_memory[dice_id - 1] + count).to_bytes(1, "little"),
+                            COMBINED_WRAM
+                        )])
+
+                    await bizhawk.write(ctx.bizhawk_ctx, [(
+                            Constants.RECEIVED_DICE_COUNT_OFFSET,
+                            len(received_dice_ids).to_bytes(1, "little"),
+                            COMBINED_WRAM
+                        )])
                 
             # Local checked checks handling
             if new_local_check_locations != self.local_checked_locations:
